/*********************************************
 * vim:sw=8:ts=8:si:et
 * To use the above modeline in vim you must have "set modeline" in your .vimrc
 * Author: Guido Socher
 * Copyright: GPL V2
 * See http://www.gnu.org/licenses/gpl.html
 *
 * This software implements a  web browser and a web server.
 * The the browser tries to access 
 * the webserver at webserver_vhost in intervals and upload measurement data.
 *
 * The on board web server can be use to read current measurements directly
 * with a web browser. It is as well used to modifiy the configuration.
 * To enter configuration mode you need to connect pin PB0 to GND for a moment.
 * You have 3 minutes to do the configuration and save it. After that the system
 * will automatically go back to the normal mode were it displays current measurement
 * values.
 * To reset the configuration to what you have hardcoded at compile time you
 * have to connect PB0 to GND for about 20 sec and then release it.
 *
 * There are 5 LEDs you can connect to the circuit to be able to
 * see what is going on.
 *
 * Chip type           : Atmega328 or Atmega644 with ENC28J60
 *********************************************/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
// http://www.nongnu.org/avr-libc/changes-1.8.html:
#define __PROG_TYPES_COMPAT__
#include <avr/pgmspace.h>
#include <avr/eeprom.h> 

#include "ip_arp_udp_tcp.h"
#include "websrv_help_functions.h"
#include "enc28j60.h"
#include "timeout.h"
#include "net.h"
#include "dnslkup.h"
#include "dhcp_client.h"
#include "analog.h"
//
// Please modify the following lines. mac and ip have to be unique
// in your local area network. You can not have the same numbers in
// two devices:
static uint8_t mymac[6] = {0x54,0x55,0x58,0x10,0x00,0x29};
// how did I get the mac addr? Translate the first 3 numbers into ascii is: TUX
//
// The name of the virtual host which you want to 
// contact (hostname of the first portion of the URL):
static char webserver_vhost[30]="ds.tuxgraphics.org";
// Our DNS resolver can only look up IP addresses in the public internet
// for testing purposes you may as well use a server in your local LAN.
// In that can you can specify just the IP address as a 
// sting in webserver_vhost.
//static const char webserver_vhost[30]="10.0.0.1";
// To disable data upload and just use the web server function you can set
// the webserver_vhost to an empty string.
//
// the URL without the hostname part:
static const char url_path[]="/cgi-bin/upld"; // this is not a buffer we change
#define ID_LEN 9
// your web-view password:
static char upload_pw[ID_LEN]="secret";
// your web-view user id:
static char upload_id[ID_LEN]="gs";
// board ID (used in case you have several boards):
static uint8_t upload_bid=1;
// default data push interval:
static uint8_t upload_interval_minutes=60;
// listen port for http on this ethernet board:
static uint16_t mywwwport=80;
// default number of ADC channels (1-8):
static uint8_t num_adc_ch=8;
//
// -- there should not be any need to change something below this line --
//
static uint8_t otherside_www_ip[4]={0,0,0,0}; // dns will fill this
// My own IP (DHCP will provide a value for it):
static uint8_t myip[4]={0,0,0,0};
// Default gateway (DHCP will provide a value for it):
static uint8_t gwip[4]={0,0,0,0};
static uint8_t arp_gw_done=0;
static uint8_t no_dns_needed=0;
#define TRANS_NUM_GWMAC 1
static uint8_t gwmac[6];
#define TRANS_NUM_WEBMAC 2
static uint8_t otherside_www_gwmac[6];
// Netmask (DHCP will provide a value for it):
static uint8_t netmask[4];
static volatile uint8_t time[3]={0,0,0}; // hour=0, min=1, sec=2
static volatile uint8_t slow_tick=0; // sec counting up and wrapping at 255
static uint8_t modify_config=0;
static uint8_t upload_minutes_cnt=1; // do one upload at the beginning, after one minute
static int8_t upload_account_state=-1; // -1 unknown, 1 no account or account expired, 0 good
//
static uint8_t upload_history_next_ptr=0; // pointer to the next element to store
static uint8_t upload_history_amount_of_data=0; // how many records we have
#define UPLOAD_HISTORY_DEPTH 12
static  uint16_t upload_history_data[UPLOAD_HISTORY_DEPTH][8];
//
#define BUFFER_SIZE 720
static uint8_t buf[BUFFER_SIZE+1];
static uint16_t gPlen;
static char gStrLarge[160]; // used by client_browse_url for the GET-string must not be touched by other functions. Only in config mode when upload is disabeled we can use it.
static char gStrbuf[22]; // The max len we need to hold is an IP addess + 1 + 4 digits for the port
static uint8_t start_web_client=0;
static uint8_t client_browse_timeout=0;
static int8_t message_state=0;
static uint16_t last_httpstatuscode=0;

// set output to VCC, red LED off
#define LEDOFF PORTB|=(1<<PORTB1)
// set output to GND, red LED on
#define LEDON PORTB&=~(1<<PORTB1)
// to test the state of the LED
#define LEDISOFF PORTB&(1<<PORTB1)
// 
// diagnostic LEDs
#define DHCP_LEDOFF PORTD|=(1<<PORTD0)
#define DHCP_LEDON PORTD&=~(1<<PORTD0)
#define DHCP_LEDISOFF PORTD&(1<<PORTD0)

#define GW_LEDOFF PORTD|=(1<<PORTD1)
#define GW_LEDON PORTD&=~(1<<PORTD1)
#define GW_LEDISOFF PORTD&(1<<PORTD1)

#define DNS_LEDOFF PORTD|=(1<<PORTD2)
#define DNS_LEDON PORTD&=~(1<<PORTD2)
#define DNS_LEDISOFF PORTD&(1<<PORTD2)

#define DATA_LEDOFF PORTD|=(1<<PORTD3)
#define DATA_LEDON PORTD&=~(1<<PORTD3)
#define DATA_LEDISOFF PORTD&(1<<PORTD3)
// 
uint16_t http200ok(uint8_t nocache)
{
        uint16_t plen;
        plen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n"));
        if (nocache){
                plen=fill_tcp_data_p(buf,plen,PSTR("Pragma: no-cache\r\n"));
        }
        return(fill_tcp_data_p(buf,plen,PSTR("\r\n")));
}

uint16_t print_line_mobile_webpage(uint16_t plen)
{
        plen=fill_tcp_data_p(buf,plen,PSTR("<meta name=viewport content=\"width=device-width\">\n"));
        return(plen);
}

uint16_t fill_tcp_data_uint16(uint8_t *buf,uint16_t plen,uint16_t i)
{
        itoa(i,gStrbuf,10); // convert integer to string
        return(fill_tcp_data(buf,plen,gStrbuf));
}

uint16_t fill_tcp_data_int8(uint8_t *buf,uint16_t plen,int8_t i)
{
        itoa(i,gStrbuf,10); // convert integer to string
        return(fill_tcp_data(buf,plen,gStrbuf));
}

// --- functions to buffer data in case of network failure
//
// we have a stack of data were we store up to 10 elements in case
// of a communication problem with the server. 
void push_upload_data_stack(void){
        int8_t i=0;
        while(i<num_adc_ch && i<8){
                upload_history_data[upload_history_next_ptr][i]=getanalogresult(i);
                i++;
        }
        upload_history_amount_of_data++;
        if (upload_history_amount_of_data>UPLOAD_HISTORY_DEPTH){
                upload_history_amount_of_data=UPLOAD_HISTORY_DEPTH;
        }
        upload_history_next_ptr=(upload_history_next_ptr+1)%UPLOAD_HISTORY_DEPTH;
}
// get the oldest data currently on the stack 
// all pointers are out-variables:
uint8_t get_upload_data_stack(uint8_t *index_to_upload_history_data){
        uint8_t i;
        if (upload_history_amount_of_data==0) return(0);
        if (upload_history_next_ptr >= upload_history_amount_of_data){
                i=upload_history_next_ptr-upload_history_amount_of_data;
        }else{
                // wrap back over zero:
                i=UPLOAD_HISTORY_DEPTH+upload_history_next_ptr-upload_history_amount_of_data;
        }
        if (i >= UPLOAD_HISTORY_DEPTH){
                return(0); // should not happen but we handle it to prevent a crash
        }
        *index_to_upload_history_data=i;
        return(1);
}
// discard oldest data currently on the stack
void shift_discard_upload_data(void){
        if (upload_history_amount_of_data) upload_history_amount_of_data--;
}
// --- end functions to buffer data in case of network failure

// read a max of maxlen char from a string stored in eeprom
uint8_t  eeprom_read_string(char *dst, uint8_t src,uint8_t maxlen){
        char c;
        uint8_t len=0;
        while((c=(char)eeprom_read_byte((uint8_t *)(src+len))) && len < maxlen){
                *dst=c;
                dst++;
                len++;
        }
        *dst='\0';
        return(len);
}

void read_eeprom_stored_data(void){
        // we have one magic number for the URL and host name and one for
        // the rest of the data
        if (eeprom_read_byte((uint8_t *)0x2) == 19){
                upload_interval_minutes=eeprom_read_byte((uint8_t *)0x3);
                mywwwport=eeprom_read_word((uint16_t *)0x4); // 4 and 5
                upload_bid=eeprom_read_byte((uint8_t *)0x6); // 6 and 7
                eeprom_read_string(upload_id,130,ID_LEN);
                eeprom_read_string(upload_pw,130+ID_LEN,ID_LEN);
        }
}

uint16_t print_webpage_config_lw(uint8_t *buf)
{
        uint16_t pl;
        pl=http200ok(0);
        read_eeprom_stored_data();
        pl=fill_tcp_data_p(buf,pl,PSTR("<h2>Modify config</h2>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<a href=/>[back]</a>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<pre><form action=/cw method=get><b>local webserver</b>\nport:<input type=text name=pn size=4 value="));
        pl=fill_tcp_data_uint16(buf,pl,mywwwport);
        pl=fill_tcp_data_p(buf,pl,PSTR(">\n\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<input type=submit value=\"set and continue\">\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<hr>\n"));
        return(pl);
}

uint16_t print_webpage_config_onlyport80(uint8_t *buf)
{
        uint16_t pl;
        pl=http200ok(0);
        pl=fill_tcp_data_p(buf,pl,PSTR("<h2>Error: can only upload to at port 80</h2>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("Do not enter a port numer in the URL, <a href=/su>[continue]</a>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<hr>\n"));
        return(pl);
}

uint16_t print_webpage_config_err(uint8_t *buf)
{
        uint16_t pl;
        pl=http200ok(0);
        pl=fill_tcp_data_p(buf,pl,PSTR("<h2>Error: data incomplete</h2>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<a href=/>[continue]</a>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<hr>\n"));
        return(pl);
}

uint16_t print_webpage_config_ok(uint8_t *buf)
{
        uint16_t pl;
        pl=http200ok(0);
        pl=fill_tcp_data_p(buf,pl,PSTR("<h2>OK</h2>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("Please reboot the board now."));
        pl=fill_tcp_data_p(buf,pl,PSTR("<hr>\n"));
        return(pl);
}

uint16_t print_webpage_config_url(uint8_t *buf)
{
        int8_t eeprom_valid=0;
        uint16_t pl;
        pl=http200ok(0);
        if (eeprom_read_byte((uint8_t *)0x1) == 18){
                eeprom_valid=1; // just for webserver_vhost and url_path
        }
        pl=fill_tcp_data_p(buf,pl,PSTR("<h2>Modify upload url</h2>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<a href=/>[back]</a>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<pre><form action=/cu method=get>\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("server: http://<input type=text name=vh value="));
        pl=fill_tcp_data(buf,pl,webserver_vhost);
        pl=fill_tcp_data_p(buf,pl,PSTR("><input type=text name=up value=\""));
        if (eeprom_valid){
                // read directly from eeprom to TCP/IP buffer without intermediate storage in a string:
                pl+=eeprom_read_string((char *)(buf + TCP_CHECKSUM_L_P+3+pl),60,50);
        }else{
                // we do not change url_path 
                pl=fill_tcp_data(buf,pl,url_path);
        }
        pl=fill_tcp_data_p(buf,pl,PSTR("\">\n(Empty all fields to disable data upload)\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<input type=submit value=\"save\">\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("</form></pre>\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<hr>\n"));
        return(pl);
}

uint16_t print_webpage_config(uint8_t *buf)
{
        int8_t i=0;
        int8_t eeprom_valid=0;
        uint16_t pl;
        uint8_t upload_intervals[]={15,30,60,120,0}; // the zero is an end marker
        pl=http200ok(0);
        if (eeprom_read_byte((uint8_t *)0x1) == 18){
               eeprom_valid=1; // just for webserver_vhost and url_path
        }
        read_eeprom_stored_data();
        pl=fill_tcp_data_p(buf,pl,PSTR("<h2>Modify config</h2>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<a href=/lw>[web serv]</a>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<pre><form action=/cc method=get>\n<b>data upload</b>\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<a href=/su>server</a>: http://"));
        pl=fill_tcp_data(buf,pl,webserver_vhost);
        if (eeprom_valid){
                // read directly from eeprom to TCP/IP buffer without intermediate storage in a string:
                pl+=eeprom_read_string((char *)(buf + TCP_CHECKSUM_L_P+3+pl),60,50);
        }else{
                pl=fill_tcp_data(buf,pl,url_path);
        }
        pl=fill_tcp_data_p(buf,pl,PSTR("\ninterval: <select name=it size=1>\n"));
        i=0;
        while(upload_intervals[i]){
                pl=fill_tcp_data_p(buf,pl,PSTR("<option value="));
                pl=fill_tcp_data_uint16(buf,pl,upload_intervals[i]);
                if (upload_interval_minutes==upload_intervals[i]){
                        pl=fill_tcp_data_p(buf,pl,PSTR(" selected"));
                }
                pl=fill_tcp_data(buf,pl,">");
                pl=fill_tcp_data_int8(buf,pl,upload_intervals[i]);
                pl=fill_tcp_data_p(buf,pl,PSTR("min</option>\n"));
                i++;
        }
        pl=fill_tcp_data_p(buf,pl,PSTR("</select>\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("id:<input type=text name=ui value="));
        pl=fill_tcp_data(buf,pl,upload_id);
        pl=fill_tcp_data_p(buf,pl,PSTR(">\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("pw:<input type=text name=pw value="));
        pl=fill_tcp_data(buf,pl,upload_pw);
        pl=fill_tcp_data_p(buf,pl,PSTR(">\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("bi:<input type=text name=bi value="));
        pl=fill_tcp_data_int8(buf,pl,upload_bid);
        pl=fill_tcp_data_p(buf,pl,PSTR(">\n\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<input type=submit value=\"save config\">\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("</form></pre>\n"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<hr>\n"));
        return(pl);
}

uint16_t print_webpage_status(uint8_t *buf)
{
        uint16_t pl;
        pl=http200ok(1);
        pl=print_line_mobile_webpage(pl);
        pl=fill_tcp_data_p(buf,pl,PSTR("<a href=/>[home]</a> <a href=/ss>[refresh]</a>\n<h2>Status info</h2>"));
        if (upload_account_state==1){
                // clear account status button if account is invalid
                pl=fill_tcp_data_p(buf,pl,PSTR("<form action=/ss method=get>\n"));
                pl=fill_tcp_data_p(buf,pl,PSTR("<input type=hidden name=ca value=1>"));
        }
        if (strlen(webserver_vhost)<3){
                pl=fill_tcp_data_p(buf,pl,PSTR("\nData upload is disabeled<br>"));
                pl=fill_tcp_data_p(buf,pl,PSTR("\n><hr>\n"));
                return(pl);
        }
        pl=fill_tcp_data_p(buf,pl,PSTR("\nData buffered due to connection error: "));
        pl=fill_tcp_data_int8(buf,pl,upload_history_amount_of_data);
        pl=fill_tcp_data_p(buf,pl,PSTR("<br>\nNext upload is in: "));
        pl=fill_tcp_data_uint16(buf,pl,upload_minutes_cnt);
        pl=fill_tcp_data_p(buf,pl,PSTR(" min<br>\nAccount status"));
        if (upload_account_state==-1){
                pl=fill_tcp_data_p(buf,pl,PSTR(": unknown"));
        }else if (upload_account_state==0){
                pl=fill_tcp_data_p(buf,pl,PSTR(": OK"));
        }else{
                pl=fill_tcp_data_p(buf,pl,PSTR(": invalid"));
                pl=fill_tcp_data_p(buf,pl,PSTR(" <input type=submit value=\"clear status and re-try\">"));
        }
        pl=fill_tcp_data_p(buf,pl,PSTR("<br>\nLast status code: "));
        pl=fill_tcp_data_uint16(buf,pl,last_httpstatuscode);
        pl=fill_tcp_data_p(buf,pl,PSTR("<br>\nGW  LED: "));
        if (GW_LEDISOFF){
                pl=fill_tcp_data_p(buf,pl,PSTR("off"));
        }else{
                pl=fill_tcp_data_p(buf,pl,PSTR("on"));
        }
        pl=fill_tcp_data_p(buf,pl,PSTR("<br>\nDNS LED: "));
        if (DNS_LEDISOFF){
                pl=fill_tcp_data_p(buf,pl,PSTR("off"));
        }else{
                pl=fill_tcp_data_p(buf,pl,PSTR("on"));
        }
        // begin dbg
        pl=fill_tcp_data_p(buf,pl,PSTR("<br>\ndbg message_state: "));
        pl=fill_tcp_data_int8(buf,pl,message_state);
        pl=fill_tcp_data_p(buf,pl,PSTR("<br>\ndbg start_web_client: "));
        pl=fill_tcp_data_int8(buf,pl,start_web_client);
        pl=fill_tcp_data_p(buf,pl,PSTR("<br>\ndbg client_browse_timeout: "));
        pl=fill_tcp_data_uint16(buf,pl,client_browse_timeout);
        // end dbg
        pl=fill_tcp_data_p(buf,pl,PSTR("<br>\n"));
        if (upload_account_state==1){
                // clear account status button if account is invalid
                pl=fill_tcp_data_p(buf,pl,PSTR("</form>\n"));
        }
        pl=fill_tcp_data_p(buf,pl,PSTR("<hr>\n"));
        return(pl);
}

uint16_t print_webpage_main(uint8_t *buf)
{
        int8_t i=0;
        uint16_t pl;
        pl=http200ok(1);
        pl=print_line_mobile_webpage(pl);
        pl=fill_tcp_data_p(buf,pl,PSTR(" <a href=/ss>[status]</a> <a href=/>[refresh]</a>\n<h2>Measurements</h2>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<pre>"));
        while(i<num_adc_ch && i<8){
                pl=fill_tcp_data_p(buf,pl,PSTR("a"));
                pl=fill_tcp_data_int8(buf,pl,i);
                pl=fill_tcp_data(buf,pl,": ");
                pl=fill_tcp_data_uint16(buf,pl,getanalogresult(i));
                pl=fill_tcp_data(buf,pl,"\n");
                i++;
        }
        pl=fill_tcp_data_p(buf,pl,PSTR("</pre>"));
        pl=fill_tcp_data_p(buf,pl,PSTR("<hr>\n"));
        return(pl);
}

// takes a string of the form command/Number and analyse it (e.g "?sw=pd7&a=1 HTTP/1.1")
// The first char of the url ('/') is already removed.
int8_t analyse_get_url(char *str)
{
        int8_t i=0;
        if (str[0]==' '){
                return(1); // end of url, main page
        }
        if (strncmp("favicon.ico",str,11)==0){
                gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 301 Moved Permanently\r\nLocation: "));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("http://tuxgraphics.org/ico/therm.ico"));
                gPlen=fill_tcp_data_p(buf,gPlen,PSTR("\r\n\r\n"));
                return(10);
        }
        // --------
        if (strncmp("ss",str,2)==0){
                if (find_key_val(str,gStrbuf,sizeof(gStrbuf),"ca")) upload_account_state=-1;
                gPlen=print_webpage_status(buf);
                return(10);
        }
        if (modify_config && strncmp("su",str,2)==0){
                gPlen=print_webpage_config_url(buf);
                return(10);
        }
        if (modify_config && strncmp("lw",str,2)==0){
                gPlen=print_webpage_config_lw(buf);
                return(10);
        }
        if (modify_config && strncmp("cu",str,2)==0){
                i=0;
                // gStrLarge is a large buffer, we need only 50
                if (find_key_val(str,gStrLarge,50,"up")){
                        i++;
                        urldecode(gStrLarge);
                }
                if (find_key_val(str,webserver_vhost,sizeof(webserver_vhost),"vh")){
                        urldecode(webserver_vhost);
                        if (strchr(webserver_vhost,':')){
                                // we can only up-load to a server at port 80:
                                gPlen=print_webpage_config_onlyport80(buf);
                                return(10);
                        }
                        i++;
                }
                if (i==2){
                        eeprom_write_block((uint8_t *)webserver_vhost,(void *)10,sizeof(webserver_vhost));
                        eeprom_write_block((uint8_t *)gStrLarge,(void *)60,50);
                        eeprom_write_byte((uint8_t *)0x1,18); // set magic number for webserver_vhost and url_path
                        return(1); // send back to config page, to really save this in eeprom one has to press save on the main config page
                }
                gPlen=print_webpage_config_err(buf);
                return(10);
        }
        if (modify_config && strncmp("cw",str,2)==0){
                if (find_key_val(str,gStrbuf,sizeof(gStrbuf),"pn")){
                        urldecode(gStrbuf);
                        gStrbuf[4]='\0'; // limit to 4 digits
                        mywwwport=atoi(gStrbuf);
                        return(1); // send back to config page, to really save this in eeprom one has to press save on the main config page
                }
                gPlen=print_webpage_config_err(buf);
                return(10);
        }
        if (modify_config && strncmp("cc",str,2)==0){
                i=0;
                if (find_key_val(str,gStrbuf,sizeof(gStrbuf),"it")){
                        urldecode(gStrbuf);
                        gStrbuf[3]='\0'; // limit to 3 digits
                        upload_interval_minutes=atoi(gStrbuf);
                        i++;
                }
                if (find_key_val(str,upload_id,ID_LEN,"ui")){
                        urldecode(upload_id);
                        i++;
                }
                if (find_key_val(str,upload_pw,ID_LEN,"pw")){
                        urldecode(upload_pw);
                        i++;
                }
                if (find_key_val(str,gStrbuf,sizeof(gStrbuf),"bi")){
                        urldecode(gStrbuf);
                        gStrbuf[3]='\0'; // limit to 3 digits
                        upload_bid=atoi(gStrbuf);
                        i++;
                }
                if (i==4){
                        eeprom_write_byte((uint8_t *)0x3,upload_interval_minutes);
                        eeprom_write_block((uint8_t *)upload_id,(void *)130,ID_LEN);
                        eeprom_write_block((uint8_t *)upload_pw,(void *)130+ID_LEN,ID_LEN);
                        eeprom_write_byte((uint8_t *)0x6,upload_bid);
                        eeprom_write_word((void *)0x4,mywwwport); // 4 and 5
                        eeprom_write_byte((uint8_t *)0x2,19); // magic number valid
                        modify_config=0;
                        gPlen=print_webpage_config_ok(buf);
                        return(10);
                }
                gPlen=print_webpage_config_err(buf);
                return(10);
        }
        return(0);
}

// called when TCNT1==OCR1A
// that is in 1Hz intervals (0.999920650393Hz intervals, clock is a bit behind)
ISR(TIMER1_COMPA_vect){
        static uint8_t dhcp_6sec=0;
        dhcp_6sec++;
        time[2]++; // step sec
        slow_tick++;
        if (dhcp_6sec>5){
                dhcp_6sec=0;
                dhcp_6sec_tick();
        }
}

// Generate an interrup about ever 1s form the 12.5MHz system clock
void timer_init(void)
{
        /* write high byte first for 16 bit register access: */
        TCNT1H=0;  /* set counter to zero*/
        TCNT1L=0;
        // Mode 4 table 14-4 page 132. CTC mode and top in OCR1A
        // WGM13=0, WGM12=1, WGM11=0, WGM10=0
        TCCR1A=(0<<COM1B1)|(0<<COM1B0)|(0<<WGM11);
        TCCR1B=(1<<CS12)|(1<<CS10)|(1<<WGM12)|(0<<WGM13); // crystal clock/1024
        // divide clock by 1024: 12.5MHz/1024=12207.0313 Hz
        // At what value to cause interrupt. You can use this for calibration
        // of the clock. Theoretical value for 12.5MHz: 12208=0x2f and 0xb0
        OCR1AH=0x2f;
        OCR1AL=0xb0;  // 0.999920650393 Hz
        // interrupt mask bit:
        TIMSK1 = (1 << OCIE1A);
}

// Status of the upload to the server on the internet
//
// the __attribute__((unused)) is a gcc compiler directive to avoid warnings about unsed variables.
void browserresult_callback(uint16_t webstatuscode,uint16_t datapos __attribute__((unused)), uint16_t len __attribute__((unused))){
        client_browse_timeout=0; // end timer (we have answer) but take out data from the buffer if the status code was 200
        if (webstatuscode==200){ // status code 200 ...
                shift_discard_upload_data();
                if (upload_history_amount_of_data>0){
                        start_web_client=1; // restart the browser and send the next from the history
                }
                upload_account_state=0;
                DATA_LEDON;
        }
        if (webstatuscode==401){ // status code 401 ...
                upload_account_state=1;
        }
        last_httpstatuscode=webstatuscode;
        LEDOFF;
}

// the __attribute__((unused)) is a gcc compiler directive to avoid warnings about unsed variables.
void arpresolver_result_callback(uint8_t *ip __attribute__((unused)),uint8_t reference_number,uint8_t *mac){
        uint8_t i=0;
        if (reference_number==TRANS_NUM_GWMAC){
                // copy mac address over:
                while(i<6){gwmac[i]=mac[i];i++;}
        }
        if (reference_number==TRANS_NUM_WEBMAC){
                // copy mac address over:
                while(i<6){otherside_www_gwmac[i]=mac[i];i++;}
        }
}

int main(void){
        uint16_t dat_p,plen;
        uint8_t days=0;
        // temporary signed and unsigend int:
        int8_t i;
        uint8_t ui;
        uint8_t dns_timeout; // wrapping 3 sec counter
        uint8_t reset_factory_default=0; // to time if the modify config button is pressed longer

        DDRB|= (1<<DDB1); // LED to indicate activity, enable PB1 as output
        LEDOFF;
        DDRD|= (1<<DDD0)|(1<<DDD1)|(1<<DDD2)|(1<<DDD3); // diga LEDs
        DHCP_LEDOFF;
        GW_LEDOFF;
        DNS_LEDOFF;
        DATA_LEDOFF;
        DDRB&= ~(1<<PINB0); // PB0 modify config, connect this to GND and then access the webserver within 2min
        PORTB|=1<<PINB0; // internal pullup resistor on
        // Set the clock speed to "no pre-scaler" (8MHz with internal osc or 
        // full external speed)
        // set the clock prescaler. First write CLKPCE to enable setting 
        // of clock the next four instructions.
        // Note that the CKDIV8 Fuse determines the initial
        // value of the CKKPS bits.
        CLKPR=(1<<CLKPCE); // change enable
        CLKPR=0; // "no pre-scaler"
        _delay_loop_1(0); // 60us

        /*initialize enc28j60*/
        enc28j60Init(mymac);
        enc28j60clkout(2); // change clkout from 6.25MHz to 12.5MHz
        _delay_loop_1(0); // 60us
        
        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        // webserver_vhost is loaded from eeprom if valid url_path is loaded when needed to save memory
        if (eeprom_read_byte((uint8_t *)0x1) == 18){
                eeprom_read_string(webserver_vhost,10,sizeof(webserver_vhost));
        }
        // other paramters:
        read_eeprom_stored_data();
        LEDON;
        timer_init();
        sei();
        adc_isr_on();
        // DHCP handling. Get the initial IP, timer needs to be running
        // for this.
        i=0;
        init_mac(mymac);
        while(i==0){
                plen=enc28j60PacketReceive(BUFFER_SIZE, buf);
                buf[BUFFER_SIZE]='\0';
                i=packetloop_dhcp_initial_ip_assignment(buf,plen,mymac[5]);
        }
        // we have an IP:
        dhcp_get_my_ip(myip,netmask,gwip); 
        client_ifconfig(myip,netmask);
        DHCP_LEDON;
        LEDOFF;

        if (strlen(webserver_vhost)<3){
                message_state=6; // disable data upload
                goto INIT_WEB_SERVER;
        }

        if (string_is_ipv4(webserver_vhost)){
                // the the webserver_vhost is not a domain name but alreay
                // an IP address written as sting
                parse_ip(otherside_www_ip,webserver_vhost);
                no_dns_needed=1;
                DNS_LEDON;
                if (route_via_gw(otherside_www_ip)==0){
                        // otherside_www_ip is in the local lan
                        arp_gw_done=1;
                }
        }
        if (gwip[0]==0 && arp_gw_done==0){
                // we must have a gateway returned from the dhcp server
                // otherwise this code will not work
                while(1); // stop here
        }
        
        if (!arp_gw_done){
                LEDON;
                // we have a gateway.
                // find the mac address of the gateway (e.g your dsl router).
                get_mac_with_arp(gwip,TRANS_NUM_GWMAC,&arpresolver_result_callback);
                while(get_mac_with_arp_wait()){
                        // to process the ARP reply we must call the packetloop
                        plen=enc28j60PacketReceive(BUFFER_SIZE, buf);
                        packetloop_arp_icmp_tcp(buf,plen);
                }
                GW_LEDON;
                LEDOFF;
        }

INIT_WEB_SERVER:
        //init the web server port. Ip and mac are the same as for the client.
        www_server_port(mywwwport);

        slow_tick=0; 
        dns_timeout=0;
        time[2]=0; // the interrupt was already running, start fresh
        while(1){
                plen=enc28j60PacketReceive(BUFFER_SIZE, buf);
                buf[BUFFER_SIZE]='\0'; // http is an ascii protocol. Make sure we have a string terminator.
                // DHCP renew IP:
                plen=packetloop_dhcp_renewhandler(buf,plen); // for this to work you have to call dhcp_6sec_tick() every 6 sec
                dat_p=packetloop_arp_icmp_tcp(buf,plen);
                if(plen==0){
                        // we are idle here
                        if (message_state==0){ // inital after power on or if upload failed
                                dns_timeout=1;
                                DNS_LEDOFF;
                                LEDON;
                                if (no_dns_needed==1){
                                        message_state=2;
                                        continue;
                                }
                                message_state=1;
                                dnslkup_request(buf,webserver_vhost,gwmac);
                                continue;
                        }
                        if (message_state==1 && dnslkup_haveanswer()){
                                dnslkup_get_ip(otherside_www_ip);
                                message_state=2;
                                continue;
                        }
                        if (message_state==2){
                                DNS_LEDON;
                                LEDOFF;
                                dns_timeout=0;
                                if (route_via_gw(otherside_www_ip)){
                                        // otherside_www_ip is behind the GW
                                        i=0;
                                        while(i<6){ otherside_www_gwmac[i]=gwmac[i]; i++; }
                                        message_state=4;
                                }else{
                                        GW_LEDOFF;
                                        LEDON;
                                        // the web server we want to contact is on the local lan. find the mac:
                                        get_mac_with_arp(otherside_www_ip,TRANS_NUM_WEBMAC,&arpresolver_result_callback);
                                        message_state=3;
                                }
                                continue;
                        }
                        if (message_state==3 && get_mac_with_arp_wait()==0){
                                // we have mac on LAN to send to
                                LEDOFF;
                                GW_LEDON;
                                message_state=4;
                        }
                        if (message_state==4){
                                message_state=5;
                                // we are ready to upload the first measurement data after startup
                                continue;
                        }
                        if (message_state==1){
                                // retry every 16 sec if dns-lookup failed:
                                if (dns_timeout > 5){
                                        message_state=0;
                                }
                                // can not put "continue" here as that would
                                // not process any code below and stop the timer
                        }
                        //----------
                        if (start_web_client==1 && modify_config==0){
                                start_web_client=2; // stop web client
                                if (get_upload_data_stack(&ui)==0){
                                        // nothing to upload
                                        client_browse_timeout=0; // stop timer
                                        continue;
                                }
                                LEDON;
                                DATA_LEDOFF;
                                if (eeprom_read_byte((uint8_t *)0x1) == 18){
                                        eeprom_read_string(gStrLarge,60,50); // url_path
                                }else{
                                        strcpy(gStrLarge,url_path);
                                }
                                strcat(gStrLarge,"?id=");
                                strcat(gStrLarge,upload_id);
                                strcat(gStrLarge,"&pw=");
                                strcat(gStrLarge,upload_pw);
                                strcat(gStrLarge,"&bi=");
                                itoa(upload_bid,gStrbuf,10);
                                strcat(gStrLarge,gStrbuf);

                                strcat(gStrLarge,"&ot=");
                                itoa((upload_history_amount_of_data - 1) * upload_interval_minutes,gStrbuf,10);
                                strcat(gStrLarge,gStrbuf);
                                strcat(gStrLarge,"&brdIp=");
                                mk_net_str(gStrbuf,myip,4,'.',10);
                                // add the port number if it is not 80:
                                if (mywwwport!=80){
                                        i=strlen(gStrbuf);
                                        gStrbuf[i]=':';
                                        itoa(mywwwport,gStrbuf+i+1,10);
                                }
                                strcat(gStrLarge,gStrbuf);
                                //  
                                // ui is set by get_upload_data_stack:
                                i=0;
                                while(i<num_adc_ch && i<8){
                                        strcat(gStrLarge,"&a");
                                        itoa(i,gStrbuf,10);
                                        strcat(gStrLarge,gStrbuf);
                                        strcat(gStrLarge,"=");
                                        itoa(upload_history_data[ui][i],gStrbuf,10);
                                        strcat(gStrLarge,gStrbuf);
                                        i++;
                                }
                                ui=strlen(gStrLarge);
                                urlencode(gStrbuf,&(gStrLarge[ui]));
                                client_browse_timeout=1; // start a timer to repeat if needed

                                // NOTE: you are no supposed to change gStrLarge until the callback function is executed
                                client_browse_url(PSTR(""),gStrLarge,webserver_vhost,&browserresult_callback,otherside_www_ip,otherside_www_gwmac);
                        }
                        // if the crystal would be exact then we
                        // would be 6.8558060448 seconds behind every day.
                        // 24*3600 - 24*3600 * 0.999920650393Hz=6.8558060448
                        // We compensate 1 sec every 4hours:
                        // 6.8558060448 - 24/4=0.8558 sec
                        // That is every day we are 0.8558 sec behind
                        // per day.
                        //
                        if (modify_config==0 && bit_is_clear(PINB,PINB0)){
                                // modify config with PB0
                                reset_factory_default=5;
                                modify_config=60; // in units of 3 sec
                        }
                        if (time[2]>59){ // sec
                                time[2]=0;
                                time[1]++; // count up min
                                if (upload_minutes_cnt) upload_minutes_cnt--;
                        }
                        if (upload_minutes_cnt==0){
                                upload_minutes_cnt=upload_interval_minutes;
                                push_upload_data_stack();
                                if (message_state==5 && upload_account_state<1) start_web_client=1; // upload one measurement every upload_interval_minutes
                                continue;
                        }
                        if (time[1]>59){ // min
                                time[1]=0;
                                time[0]++; // count up hour
                                // We calibrate the clock every 4hours a bit:
                                // our clock is a bit behind.
                                if (time[2]%4==0) time[2]+=1; // adjust/compensate
                        }
                        if (time[0]>23){ // hours
                                time[0]=0;
                                days++;
                        }
                        // every 10 days:
                        if (days==10){
                                time[2]+=8; // adjust/compensate 
                                days=0; // reset to zero otherwise we would do +clock_cal for a whole day
                        }
                        if (slow_tick > 3){
                                slow_tick=0;
                                // every 3 sec:
                                if (client_browse_timeout) client_browse_timeout++;
                                if (client_browse_timeout>4){
                                        if (message_state==5) message_state=0; // redo the dns lookup
                                        client_browse_timeout=0;
                                        // now the upload will happen at the next interval
                                }
                                if (dns_timeout) dns_timeout++;
                                if (modify_config) modify_config--;
                                if (modify_config) {
                                        // check if somebody is holding the button for a long time
                                        if (bit_is_clear(PINB,PINB0)){
                                                reset_factory_default++;
                                        }else{
                                                // hold PB0 to GND for 15 sec to reset data to factory default
                                                if (reset_factory_default>10){
                                                        // button now released, set modify_config to zero in 3 sec 
                                                        // This is just to handle any button bouncing and not to 
                                                        // enable modify_config immediately again.
                                                        modify_config=1;
                                                        eeprom_write_byte((uint8_t *)0x1,1); // delete magic number
                                                        eeprom_write_byte((uint8_t *)0x2,1); // delete magic number
                                                }else{
                                                        // did not hold the PB0 pin to GND long enough
                                                        reset_factory_default=0;
                                                }
                                        }
                                }
                        }
                        continue;
                }
                if(dat_p==0){ // plen!=0 and dat_p==0 means some packet but not TCP
                        // check for incomming messages not processed
                        // as part of packetloop_arp_icmp_tcp, e.g udp messages
                        udp_client_check_for_dns_answer(buf,plen);
                        continue;
                }
                // when dat_p!=0 then we get normally a http  request 
                if (strncmp("GET ",(char *)&(buf[dat_p]),4)!=0){
                        // head, post and other methods:
                        //
                        // for possible status codes see:
                        // http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html
                        gPlen=http200ok(0);
                        gPlen=fill_tcp_data_p(buf,gPlen,PSTR("<h1>200 OK</h1>"));
                        goto SENDTCP;
                }
                // the +5 takes off the first "/"
                i=analyse_get_url((char *)&(buf[dat_p+5]));
                if (i==-1){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 401 Unauthorized\r\nContent-Type: text/html\r\n\r\n<h1>401 Unauthorized</h1><a href=/>continue</a>"));
                        goto SENDTCP;
                }
                if (i==0){
                        gPlen=fill_tcp_data_p(buf,0,PSTR("HTTP/1.0 404 Not Found\r\nContent-Type: text/html\r\n\r\n<h1>404 Not Found</h1>"));
                        goto SENDTCP;
                }
                if (i==10){
                        goto SENDTCP;
                }
                // just display the status:
                // normally i==1
                if (modify_config >0 ){
                        gPlen=print_webpage_config(buf);
                }else{
                        gPlen=print_webpage_main(buf);
                }
                //
SENDTCP:
                www_server_reply(buf,gPlen); // send web page data
                // tcp port www end
        }
        return (0);
}
